import pytest
import jpcm
import matplotlib.pyplot as plt
import matplotlib.cm.listedcolormap as LCM

def test_jpcm_load():
    """
    Test that the jpcm package can be loaded.
    """
    assert jpcm

def test_jpcm_get():
    """
    Test that the jpcm package can be loaded.
    """
    assert isinstance(jpcm.get("def"),LCM)

def test_jpcm_register():
    """
    Test that the jpcm package can be loaded.
    """
    assert jpcm.register()



#!/usr/bin/env python3

"""
jpcm | perceptual-jp-colormaps  
  
A perceptually uniform colormap generator for Matplotlib equipped with traditional-ish Japanese colors to serve as a fixed color palette.
"""
from matplotlib.colors import ListedColormap as LCM
import jpcm.core.core as core
import jpcm.load.load as load
import jpcm.core.maps as maps

def register(custom_maps=None,datafile=None):
    """
    registers default and custom colormaps (if given), and saves to datafile (is given)
    args: custom_maps (dict) : colormap dictionary of keycolors, datafile (str) : file path to save at 
    """
    cmaps = {}
    cmaps.update(maps.cmaps)
    if datafile is not None: cmaps.update(custom_maps)
    mapdata = core.gen_cmaps(cmaps, memory_only = True)
    load.jpcm_register(mapdata)
    if datafile is not None: core.save(mapdata,datafile)

def open(datafile = f'{maps.path}cmaps.txt'):
    """
    registers colormaps given by datafile
    args: datafile (str) : file path containing colormap data
    """
    load.jpcm_register(load.jpcm_load(datafile))

#mainly for MATPLOTLIB 3.3/3.4
def get(key,segmented = False,cmaps=maps.cmaps):
    RGB,_ = core.get(key,segmented = False,cmaps=maps.cmaps)
    return LCM(RGB.tolist())


import numpy as np

segmentOPT="_segmented"
path0 = "maps/"
path1 = f'src/jpcm/{path0}'
path = f'../{path0}'
suffix = ".png"

def rgb(r,g,b,a=1):
    arr = np.array([r,g,b])/255
    if a != 1 :
        return [arr[0],arr[1],arr[2],a]
    return arr

# list of colors (skytree white not accurate)
transparent = rgb(0,0,0,0)
# whites
aijiro = rgb(235, 246, 247)
aijiro_alpha = rgb(235, 246, 247,0.15)
sukaitsuri_iro = rgb(220, 240, 250)
mashiro = rgb(255, 255, 255)
zoge_iro = rgb(255, 255, 240)
hakushi = rgb(249, 251, 255)
# blue-greens
mizu_iro = rgb(134, 171, 165)
gunjo_iro = rgb(93, 140, 174)
sora_iro = rgb(77, 143, 172)
chigusa_iro = rgb(49, 117, 137)
benimidori = rgb(120, 119, 155)
kurobeni = rgb(35, 25, 30)
ruri_iro = rgb(58,36,59)
rurikon = rgb(27, 41, 75)
konjo_iro = rgb(0, 49, 113)
aoi = rgb(0, 0, 200)
sabi_asagi = rgb(106,127,122)
mizu_asagi = rgb(116,159,141)
omeshi_onando = rgb(61, 76, 81)
tetsuonando = rgb(43, 55, 54)
mo_egi = rgb(91,137,48)
seiji_iro = rgb(129,156,139)
# purples
shikon = rgb(43,32,40)
kokimurasaki = rgb(58,36,59)
kuwazome = rgb(89,41,44)
murasaki = rgb(79,40,75)
ayame_iro = rgb(118,53,104)
# reds
sakuranezumi = rgb(172, 129, 118)
ginshu = rgb(188, 45, 41)
azuki_iro = rgb(103, 36, 34)
akabeni = rgb(195,39,43)
shinshu = rgb(143,29,33)
karakurenai = rgb(201,33,55)
enji_iro = rgb(157,41,51)
# browns
benihibata = rgb(111,48,40)
# pinks
nakabeni = rgb(201,55,86)
# yellows
tomorokoshi_iro = rgb(250,169,69)
# greys
haiiro = rgb(91, 90, 92)
# blacks
sumi_iro = rgb(39,34,31)
kokushoku = rgb(23,20,18)
# list of colormaps

cmaps = \
{
    'def' : [rurikon,chigusa_iro,benimidori,azuki_iro],
    # 'haru' : [],    # warm, bright light : orange, yellow, coral pink, warms
    # 'natsu' : [],   # cool, muted light  : blue-based pink, soft purple, greyed blues, pastels
    # 'aki' : [],     # warm, muted dark   : warm browns, olive/moss greens, teracotta orange, muted reds
    'fuyu' : [aijiro,gunjo_iro,mizu_iro,akabeni,murasaki,kokushoku],    # cool, bright dark  : bright red, dark pink, black, white
    # 'printemps' : [],
    # 'ete' : [],
    # 'automne' : [],
    # 'hiver' : [],
    'ice'   : [aijiro,gunjo_iro],
    'iron-ice' : [sabi_asagi,omeshi_onando],
    'water' : [gunjo_iro,mizu_iro,seiji_iro],
    'momiji': [sakuranezumi,ginshu,azuki_iro,enji_iro],
    'sky' : [kokushoku,kokimurasaki,sora_iro,aijiro],
    # inspired by cmasher (https://cmasher.readthedocs.io/)
    'sunburst' : [kokushoku,enji_iro,akabeni,tomorokoshi_iro,zoge_iro],
    'flamingo' : [kokushoku,shinshu,nakabeni,aijiro],
    'tree'     : [kokushoku,benihibata,mo_egi]

}


#!/usr/bin/env python3
# inspired / based on https://stackoverflow.com/questions/61487041/more-perceptually-uniform-colormaps
import matplotlib
matplotlib.use('agg')
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap as LCM
from matplotlib.colors import Normalize as mplnorm
import matplotlib.cm as cm
import logging
import colour
colour.utilities.filter_warnings(*[True] * 4)
import numpy as np
import json as js
from scipy.interpolate import CubicSpline,interp1d
from pynverse import inversefunc
# from savitzky_golay import savitzky_golay
import jpcm.core.maps as maps
import jpcm.core.readme as readme

prefix = "jp-cm-"
defaultdatafile = f'{maps.path}cmaps.txt'
discretization = 40

baseColorSpace = "sRGB"
colorModel = 'CAM16UCS' # ['CAM16UCS','Jzazbz',...] # 
outputColorModel = 'sRGB'

# spline_type= 'cubic' # ['custom','cubic']
spline_mode = 'natural' # ['natural','clamped','not-a-knot'] (in best-to-worst order)

n_iterations = 1 # number of adjustments # note >1 values cause divergence!

# logging.basicConfig(filename=f'{maps.path}log.txt')
# logging.getLogger('matplotlib.font_manager').disabled = True
# logger = logging.getLogger(__name__)
# logger.level=logging.DEBUG

def linear_segmented_spline(x):
    """
    Creates segmented linear spline between color points.
    """
    n_gradients = len(x) - 1
    RGB_list = []
    for i in range(n_gradients):
        gradient = colour.utilities.lerp(
            x[i][np.newaxis],
            x[i+1][np.newaxis],
            np.linspace(0, 1, discretization)[..., np.newaxis])
        RGB_list.append(gradient)
    return np.vstack(RGB_list)

def smooth_spline(x):
    """
    Creates a cubic spline between color points,
    then adjusts to get a (mostly) perceptually uniform colormap.
    """
    n_keypoints = len(x)
    n_points = (n_keypoints-1)*discretization
    naive_xs = np.linspace(0,n_keypoints-1,n_points)

    spline = CubicSpline(list(range(n_keypoints)), x, bc_type=spline_mode)
    naive = spline(naive_xs)

    adjusted = naive
    adjusted_xs = naive_xs
    for _ in range(n_iterations):
        adjusted_xs = adjust_spline(adjusted, adjusted_xs, n_keypoints)
        adjusted = spline(adjusted_xs)

    # return np.array(
    # [savitzky_golay(adjusted.T[0], 40, 5),
    # savitzky_golay(adjusted.T[1], 40, 5),
    # savitzky_golay(adjusted.T[2], 40, 5)]).T

    return adjusted

def adjust_spline(naive,naive_xs, n_keypoints):
    """
    Adjust spline speed along curve to force perceptual derivative to zero.
    """
    _,local_derivs = fitD(naive)
    difficulty = np.cumsum(local_derivs)
    stepDiff = difficulty[len(difficulty)-1] / (n_keypoints-1)
    diffFunction = interp1d(naive_xs[1:], difficulty,fill_value="extrapolate")
    return inversefunc(diffFunction,y_values = naive_xs * stepDiff)

def delta_ymax(values):
    return max(np.max(values)*1.1, 0)

def fitD(gradient):
    """
    (Lifted from https://github.com/1313e/viscm/blob/master/viscm/gui.py)
    """
    local_deltas = np.sqrt(np.sum(np.diff(gradient, axis=0)**2, axis=-1))
    local_derivs = (len(local_deltas)-1)*local_deltas
    return local_deltas,local_derivs

def colormap(gradient, RGB, name):
    """
    Calculate and plot colormap with perceptual derivative. Returns matplotlib colormap.
    """
    n_keys = int(len(gradient)/discretization)

    fig,_ = colour.plotting.plot_multi_colour_swatches(
        [colour.plotting.ColourSwatch(RGB=np.clip(x, 0, 1)) for x in RGB], height = 2*n_keys)

    ax = fig.add_subplot(212)  

    cmap = LCM(RGB,name=name)

    local_deltas,local_derivs = fitD(gradient)

    ax.plot(local_derivs,c=maps.ginshu)
    arclength = np.sum(local_deltas)
    rmse = np.std(local_derivs)
    ax.text(0.1, 0.4, f"Perceptual Derivative for Colormap : {name}", transform=ax.transAxes)

    ax.text(0.1,0.1,"Length: %0.1f\nRMS deviation from flat: %0.1f (%0.1f%%)"
            % (arclength, rmse, 100*rmse / arclength),transform=ax.transAxes)
    # print("Perceptual derivative: %0.5f +/- %0.5f" % (arclength, rmse))
    ax.set_ylim(0, delta_ymax(local_derivs))
    ax.get_xaxis().set_visible(False)
    ax.margins(0,0)
    ax.set_facecolor(maps.aijiro_alpha)
    fig.patch.set_facecolor(maps.transparent)
    ax.set_aspect(n_keys*8/delta_ymax(local_derivs))
    return cmap,fig

def draw(gradient, RGB, name):
    # fig = plt.figure(constrained_layout=True, figsize=(10, 4))
    # fig.patch.set_facecolor(maps.transparent)
    # subfigs = fig.subfigures(1,2, wspace=0.00, hspace=0.00,width_ratios=[2, 1])
    # for subfig in subfigs : subfig.set_facecolor(maps.transparent)
    cmap,fig = colormap(gradient, RGB, name)
    ax3 = fig.add_axes([1.0, 0.1, 0.5, 0.5], projection="3d")
    ax3.set(xlim3d=(0, 1), xlabel='L')
    ax3.set(ylim3d=(-0.5, 0.5),ylabel='M')#ylim3d=(0, 1),
    ax3.set(zlim3d=(-0.5, 0.5), zlabel='S') # zlim3d=(0, 1),
    ax3.scatter(gradient[:,0],gradient[:,1],gradient[:,2],c=list(range(gradient.shape[0])),cmap=cmap)
    ax3.set_facecolor(maps.transparent)
    ax3.xaxis.pane.fill = False
    ax3.yaxis.pane.fill = False
    ax3.zaxis.pane.fill = False
    ax3.xaxis.pane.set_edgecolor(maps.aijiro_alpha)
    ax3.yaxis.pane.set_edgecolor(maps.aijiro_alpha)
    ax3.zaxis.pane.set_edgecolor(maps.aijiro_alpha)
    return fig

def get(key,segmented = False,cmaps=maps.cmaps):
        x = colour.convert(cmaps[key], 'Output-Referred RGB', colorModel)
        gradient = linear_segmented_spline(x) if segmented else smooth_spline(np.array(x))
        RGB = np.array([np.clip(x, 0, 1) for x in colour.convert(gradient, colorModel, outputColorModel)])
        return RGB, gradient

def gen_cmaps(cmaps,memory_only = True):
    """
    Create colormaps and save individual image as png, and all maps to txt.
    args:   map (dict) : colormap dictionary, memory_only (bool) : save flag for colormap details
    out:    ../maps/maps.txt
            ../maps/<name>.png
    """
    

    mapdata = {}

    for i in range(2):
        segmented = (i==0)
        options = maps.segmentOPT if segmented else ""

        for key in cmaps.keys():

            name = prefix+key+options

            RGB,gradient = get(key,segmented,cmaps)

            # logger.debug("Perceptual:{}".format(x))
            # logger.debug("RGB:{}".format(RGB))

            if not memory_only: 
                fig = draw(gradient, RGB, name)
                fig.savefig(maps.path+key+options+".png",bbox_inches='tight')
                plt.close()

            mapdata[name] = RGB.tolist()
       

    return mapdata

def save(mapdata,datafile=defaultdatafile):
    with open(datafile, 'w') as file:
        file.write(js.dumps(mapdata))

def core():
    save(gen_cmaps(maps.cmaps, memory_only = False))
    readme.generate()
    # gen_cmaps(cmaps,("--segmented" in sys.argv))

if __name__=='__main__':
    core()



from jpcm.core.maps import *
readmefile = "/../../../../README.md"
uname='akhilsadam'
rname='perceptual-jp-colormaps'
branch='master'
readme = """
# perceptual-jp-colormaps  
  
A perceptually uniform colormap generator for Matplotlib equipped with traditional-ish Japanese colors to serve as a fixed color palette.   
(Culture appropriation at its finest.) Additions are welcomed.  

Why perceptually uniform colormaps? Can you not just use default Matplotlib colormaps?   
- non-perceptually uniform colormaps induce dangerous artifacts, as seen in the below image: ![top-view of pyramid](https://i.stack.imgur.com/JcTDb.png).
The left colormap introduces new features to the data.
- Matplotlib perceptually uniform colormaps are not visually appealing and sometimes lack enough contrast.
With this package, anyone can design colormaps to fit their visual style.

dependencies:  
- Python3:  
--	colour-science  
--	numpy  
--	logging  
--	scipy  
--	json  
--	pynverse  
--	matplotlib  
  
tested on:  
- Windows 10  
  
expected to work on:  
- Linux  
- MacOS  
  
install:
- `pip3 install jpcm` or download source

run generator:   
- to create your own colormaps 
   - as a package:  
       - `import jpcm` 
       - `jpcm.register()` will register all default colormaps and any additional ones via the optional `custom_maps` argument. 
        Note any custom_maps should be of the following format (note the RGB key colors):  
            `cmaps = {  
                'def' : [[0,0,0],[255,0,0]],  
            }`  
       - The optional `datafile` argument will cause jpcm to save all colormaps to that location as well.
  - directly via code:
       - edit maps.py with your preferred cmap styles, and run `python3 gen.py` or `python3 gen.py --readme` from `/gen/`

load colormaps:
- `import jpcm`  
- `jpcm.open(<filename>)` will return a dictionary containing all the colormaps in the file at <filename>. 
- Note the file should be of the same type as the generated 'cmap.txt' from the generator


## gallery  
"""

def generate():
    lines = []
    for key in cmaps.keys():
        names = [path1+key+suffix,path1+key+segmentOPT+suffix]
        imagenames = ["https://github.com/{}/{}/blob/{}/{}?raw=true".format(uname,rname,branch,name) for name in names]
        line = "\n![]("+"{}".format(imagenames[0])+") ![]("+"{}".format(imagenames[1])+")"
        lines.append(line)
    readmeHT = readme + "  ".join(lines)  
    with open(readmefile, 'w') as file:
        file.write(readmeHT)



import matplotlib
from matplotlib.colors import ListedColormap as LCM
import matplotlib as mpl
import numpy as np
import json as js

rev = "_rev"

def jpcm_load(datafile):

    with open(datafile,'r') as file:
        mapdata = js.load(file)

    return {key: (np.array(mapdata[key]),key) for key in mapdata.keys()}

def jpcm_register(jpcm_cmaps):
    for key in jpcm_cmaps.keys():
        if key not in matplotlib.pyplot.colormaps():
            mpl.colormaps.register(LCM(jpcm_cmaps[key]),name = key)
            # mpl.colormaps.register(np.flip(LCM(jpcm_cmaps[key]),axis=0),name = key+rev)


